/*
 * This program is Copyright 1996 Mark Eichin, and is freely redistributable
 * under the terms of the GNU General Public License. See the file COPYING for
 * further details.
 */
#include <termios.h>
#include <unistd.h>
#include <stdio.h>

char *keys[3][4] = {
  { "SP", "DEL", "BS", "ENT", },
  { "e", "f", "g", "h", },
  { "a", "b", "c", "d", },
};

#define MAPKEY(tw,sl) ((tw >> (sl+sl)) & 3)
#define GKEYS(tw,sl) keys[MAPKEY(tw,sl)-1][sl]

char *a2xmods[] = {
  "\024\023",			/* shift: c-t c-s */
  0,				/* num -- don't fall back*/
  0,				/* fn -- don't fall back*/
  "\024\003",			/* ctl: c-t c-c */
  "\024\015",			/* alt(meta): c-t c-m */
  0,				/* mouse already dealt with */
};

/* If we just index off the finger bits, and treat the thumb as a
   modifier only, we get 4^4 (256) array elements -- not worth
   searching, just keep an array of pointers (since it's only 1K.)
   However, we only want "just send the modifier" to be a default, not
   the only action; this gives us 6 times the storage (5 modifiers and
   the original table.) Since the program text is already 80K (and why
   is that?) I'm not going to sweat it... */
/* no, the modifiers can be used in combination... even ignoring 
   "mouse" that leaves 2^5 or 32K of table space. Hmm. Oh well, just
   do it for now and optimize it later. */
/* Hmm, only shift, ctl, and alt even have prefixes; how about always
   failing for the others (fn and num) rather than emit something 
   unexpected? */

/* so: look up key+all mods. If we don't find it, we could
   -- strip them off one at a time, adding the simple modifier, and
      looking it up again. (This could be ambiguous...)
   -- strip them *all* off, and add the modifiers?
   */

/* there should be an array of pointers to mod tables, since that way
   we can avoid using any storage much of the time */

typedef char *modtab[256];
modtab *modtables[32];
/* 0 is inherently "none" anyway, so just use the bits as index
   directly */

int a2xcvt(char *buf, int tw)
{
  int i;
  modtab *thismodtable = 0;
  char *thistext = 0;
  int thisthumb;

  *buf = 0;

  thisthumb = (tw >> 8) & 31;
  thismodtable = modtables[thisthumb];
  if (thismodtable)
    thistext =  (*thismodtable)[tw & 255];

  if (!thistext) {
    thistext =  (*modtables[0])[tw & 255];
    if (!thistext)
      /* there wasn't anything to fall back to */
      return 0;

    /* try falling back */
    if (thisthumb & ( (1<<1) | (1<<2) ))
      /* NUM or FN -- don't */
      return 0;

    for (i = 0; i < 5; i++)	/* 5 -- not Mouse */
      if (thisthumb & (1<<i))
	if (a2xmods[i])
	  strcat(buf, a2xmods[i]);
  }
  strcat(buf, thistext);
  return strlen(buf);
}

int a2xmouse(char *buf, int tw)
{
  int i;
  modtab *thismodtable = 0;
  char *thistext = 0;
  int thisthumb;

  *buf = 0;

  thisthumb = (tw >> 8) & 31;
  thismodtable = modtables[thisthumb];
  if (thismodtable)
    thistext =  (*thismodtable)[tw & 255];

  if (!thistext) {
    thistext =  (*modtables[0])[tw & 255];
    if (!thistext)
      /* there wasn't anything to fall back to */
      return 0;

    /* try falling back */
    if (thisthumb & ( (1<<1) | (1<<2) ))
      /* NUM or FN -- don't */
      return 0;

    for (i = 0; i < 5; i++)	/* 5 -- not Mouse */
      if (thisthumb & (1<<i))
	if (a2xmods[i])
	  strcat(buf, a2xmods[i]);
  }
  strcat(buf, thistext);
  return strlen(buf);
}

/* TIOCMGET puts back get_modem_info */
/* In theory, using termios to set the baud rate to zero should have */
/* the side effect of dropping DTR. Look it up first, though. */
/* Assumes stdin is the control channel! */

#if 0
/* OLD linux-specific version */
int
setup_tty()
{
  int st;
  int gmi = 0;

  system("stty 2400 raw");

  st = ioctl(0,TIOCMGET, &gmi);
  if(st) perror("tiocmget");
  gmi &= ~TIOCM_DTR;
  gmi |= TIOCM_RTS;
  st = ioctl(0,TIOCMSET, &gmi);
  if(st) perror("tiocmset");
}
#else
/* this should be much more portable, although it has only been tested 
   on linux */
int
setup_tty()
{
  struct termios t;

  tcgetattr(0, &t);
  cfsetispeed(&t, B2400);
  tcsetattr(0, TCSANOW, &t);
  cfsetispeed(&t, B0);
  tcsetattr(0, TCSANOW, &t);
}
#endif

int yywrap() {
  /* indicate to lex that no more input is available. */
  return 1;
}
extern FILE* yyin;

int
main(int argc, char** argv) {
  unsigned char buf[128];
  FILE* parserfile;

  parserfile = fopen(argv[1], "r");
  if(!parserfile) {
    fprintf(stderr,"%s: couldn't open parser file <%s>\n", argv[0], argv[1]);
    exit(1);
  }
  yyin = parserfile;

  setup_tty();

  fprintf(stderr, "parse done: %d\n",  yyparse());

  while(1) {
    unsigned char c;
    static enum { START, SECOND, MOUSE1, MOUSE2, MOUSE3, OTHER } state = START;
    static unsigned char b1 = 0, b2 = 0;
    static int klast = 0, kthis = 0;
    static enum { PUSHING, RELEASING } st2 = PUSHING;
    static int mlast = 0, mbits = 0, mlr = 0, mfb = 0;
    static int mhuh = 0;
    static enum { MZERO, MONEDN, MONEUP, MMOVING, MTWODN } mstate = MZERO;


    read(0,buf,1);
    c = buf[0];
    if (0 == (c & 128))  {
      state = START;
    }
	
    switch (state) {
    case START:
      {
	b1 = c;
	state = SECOND;
      }
      break;
    case SECOND:
      {
	b2 = c;
	
	kthis = ( (b2 & 0x7f) << 7) | b1;
	
	if (klast != kthis) {
	  switch (st2) {
	  case PUSHING:
	    if (klast < kthis) { 
	      klast = kthis;
	    } else {
	      char xbuf[1024];
	      int xlen;
	      if (mstate == MMOVING)
		xlen = a2xmouse(xbuf, klast);
	      else
		xlen = a2xcvt(xbuf, klast);
	      write(1,xbuf,xlen);
	      st2 = RELEASING;
	    }
	    break;
	  case RELEASING:
	    if (klast > kthis) {
	      /* still releasing */
	    } else {
	      st2 = PUSHING;
	    }
	    klast = kthis;
	  }
	}
	state = MOUSE1;
      }
      break;
    case MOUSE1:
      {
	c &= 127;
	mfb = mlr = mbits = 0;
	mbits |= c << 14;
	mfb = c << 1;
	state = MOUSE2;
      }
      break;
    case MOUSE2:
      {
	c &= 127;
	mbits |= c << 7;
	mlr |= c>>1;

	mfb |= (c & 3) << 8;
	state = MOUSE3;
      }
      break;
    case MOUSE3:
      {
	c &= 127;
	mbits |= c;
	mlr |= c << 6;
	state = OTHER;
	switch (mstate) {
	case MZERO:
	  if (kthis & (1<<13)) mstate = MONEDN;
	  break;
	case MONEDN:
	  if (!(kthis & (1<<13))) mstate = MONEUP;
	  break;
	  /* MONEUP -> MMOVING done after acquiring the value */
	case MMOVING:
	  if (kthis & (1<<13)) mstate = MTWODN;
	  break;
	case MTWODN:
	  if (!(kthis & (1<<13))) mstate = MZERO;
	  break;
	}

	if (mstate == MMOVING || mstate == MONEUP) {
	  char xb[100];
	  /* this takes 10bit values to 8bit, with signs */
	  char clr = mlr >> 2, cfb = mfb >> 2; 
	  static omlr=0, omfb=0;

	  static int orig_x = 0, orig_y = 0;
	  static int last_x = 0, last_y = 0;
#if 0
fprintf(stderr,"clr %d cfb %d o(%d,%d) l(%d,%d)\n\r",
	clr, cfb, orig_x, orig_y, last_x, last_y);
#endif

	  switch (mstate) {
	  case MONEUP:
	    orig_x = clr;
	    orig_y = cfb;
	    mstate = MMOVING;
	    omlr = mlr;
	    omfb = mfb;
	  }
#define iabs(x) ( ((x)>0)?(x):(-(x)))

	  if ((iabs(omlr-mlr) > 4) || (iabs(omfb-mfb) > 4)) {
	    omlr = mlr;
	    omfb = mfb;
	  if (last_x != clr || last_y != cfb) {
#if 0
fprintf(stderr,"moving by: (%d,%d)\n\r",3*(orig_x-clr), 5*(orig_y-cfb));
#endif
	    sprintf(xb, "\024\004%d %d\024", 3*(orig_x-clr), 5*(orig_y-cfb));
	    write(1,xb,strlen(xb));
	  }
	  }
	  last_x = clr; last_y = cfb;
	  orig_x = clr; orig_y = cfb; /* normal mode, not xtalq */
	}
	mlast = mbits;
      }
      break;
    case OTHER:
      {
	/* extra bytes? ignore them. */
      }
      break;
    default:
      exit(1);
      break;
    }
  }
}
