/* Parse the twiddler.ini file as described in the twiddler docs */
/* Supposedly the RHS is like a Microsoft/MouseSystems "TYPE" file... */
/*
 * This program is Copyright 1996 Mark Eichin, and is freely redistributable
 * under the terms of the GNU General Public License. See the file COPYING for
 * further details.
 */

/* file format:
   LINE := THUMB WSP FINGER EQUALS TEXT
   THUMB := "0" | MOD { "+" MOD }
   MOD :=   "shift"| "num"| "func"| "ctrl"| "alt"
   WSP := { space | tab }
   FINGER := DIGIT DIGIT DIGIT DIGIT
   DIGIT := "0" | "L" | "M" | "R"
   EQUALS := WSP "=" WSP
   TEXT := TEXTTOKEN { "," TEXTTOKEN }
   TEXTTOKEN := QUOTEDTEXT | NUMBER | KEYNAME | KEYCODE
   QUOTEDTEXT := """ ANYCHAR """
   NUMBER := 0..9 { 0..9 }
   KEYNAME := "BACKSPACE" | "ENTER" | "ESC" | "TAB"
   KEYCODE := "0," NUMBER
*/
/* yacc decls */
%{
typedef struct {
  int thumb_bits;
  int finger_bits;
  char *text;
} grip_rec;

typedef struct {
  int thumb;
  char *name;
} mod_value;
#include <stdio.h>
int curr_line = 1;
/* #define YYDEBUG 1 */
#undef yywrap
%}

%union {
  grip_rec grip;
  int thumb;
  int finger;
  char *string;
  mod_value mod;
  int keyval;
  int number;
}
%token NL
%token WSP
%token QUOTEDTEXT
%token DECIMAL
%token ZERO

%token MOD_SHIFT
%token MOD_NUM
%token MOD_FUNC
%token MOD_CTRL
%token MOD_ALT
%token MOD_MOUSE
%token KEYNAME_BSP
%token KEYNAME_ENTER
%token KEYNAME_ESC
%token KEYNAME_TAB
%%
/* grammar */

input : /* empty */
	| input line
;

line : NL
	| realline NL {  push_keys($1.grip); }
        | error  NL { yyerrok; }
;

realline: thumb { /* printf("got thumb\n"); */} 
	  WSP
	  finger {  /* printf("got finger\n"); */}
	  equals
	  text   {  /* printf("got text <%s>\n",$7.string); */ }
	       { $$.grip.thumb_bits = $1.thumb;
		 $$.grip.finger_bits = $4.finger; 
		 $$.grip.text = $7.string;
	       }
;

thumb : ZERO { $$.thumb = 0; }
	| thumbs
;

thumbs : mod { $$.thumb = $1.mod.thumb; }
	| thumbs '+' mod { $$.thumb = $1.thumb | $3.mod.thumb; }
;

finger : digit digit digit digit
	 { $$.finger = $1.finger + ($2.finger<<2)  + ($3.finger<<4)
	     + ($4.finger<<6); }
;

digit : ZERO   { $$.finger = 0; /* printf("got ZERO digit\n"); */}
	| 'l'  { $$.finger = 1; }
	| 'm'  { $$.finger = 2; }
	| 'r'  { $$.finger = 3; }
;

equals : WSP '=' WSP
;

text :	texttoken { $$.string = stralloc($1.string); /* printf("text <%s>\n", $$.string); */ strfree($1.string); }
	| text ',' WSP texttoken { $$.string = strappend($$.string,$4.string); strfree($4.string); }
	| text ',' texttoken { $$.string = strappend($$.string,$3.string); strfree($3.string); }

texttoken : QUOTEDTEXT { $$.string = $1.string; }
	| number { $$.string = malloc(1); sprintf($$.string,"%c",$1.number); }
	| keyname { $$.string = malloc(1); sprintf($$.string,"%c",$1.keyval); }
;

number : DECIMAL { $$.number = $1.number; }
	| ZERO { $$.number = $1.number; }
	| number DECIMAL { $$.number = $1.number*10 + $2.number; }
	| number ZERO { $$.number = $1.number*10 + $2.number; }
;

mod : 	MOD_SHIFT	{ $$.mod.thumb = 1 << 0; }
	| MOD_NUM	{ $$.mod.thumb = 1 << 1; }
	| MOD_FUNC	{ $$.mod.thumb = 1 << 2; }
	| MOD_CTRL	{ $$.mod.thumb = 1 << 3; }
	| MOD_ALT	{ $$.mod.thumb = 1 << 4; }
	| MOD_MOUSE	{ $$.mod.thumb = 1 << 5; }
;

keyname : KEYNAME_BSP	{ $$.keyval = 8; }
	| KEYNAME_ENTER	{ $$.keyval = 13; }
	| KEYNAME_ESC	{ $$.keyval = 27; }
	| KEYNAME_TAB	{ $$.keyval = 9; }
;
%%
/* other C code */

/* MOD :   "shift"| "num"| "func"| "ctrl"| "alt" */
/* WSP : { space | tab } */
/* QUOTEDTEXT : '"' ANYCHAR '"'*/
/* KEYNAME : "BACKSPACE" | "ENTER" | "ESC" | "TAB" */
/* DECIMAL : '0'..'9', .number = 0..9 */
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

strfree(s)
char* s;
{
  free(s);
}

char *stralloc(s)
     char *s;
{
  int l = strlen(s)+1;
  char *r = malloc(l);
  strcpy(r,s);
  return r;
}

char *strappend(s,t)
     char *s;
     char *t;
{
  int l = strlen(s) + strlen(t) + 1;
  char *r = realloc(s,l);
  strcat(r,t);
  return r;
}

typedef char *modtab[256];
modtab *modtables[32];

push_keys(gr)
     grip_rec gr;
{
  modtab *thismodtable;
  int i;

  fprintf(stderr, "grip: thumb 0x%x finger 0x%x text <%s> [%d]\n",
		  gr.thumb_bits, gr.finger_bits, gr.text, strlen(gr.text));

  thismodtable = modtables[gr.thumb_bits];
  if (!thismodtable) {
    fprintf(stderr, "grip: new thumb table %d\n", gr.thumb_bits);
    modtables[gr.thumb_bits] = thismodtable = (modtab*)malloc(sizeof(modtab));
    for (i = 0; i < 256; i++) (*thismodtable)[i] = 0;
  }
  (*thismodtable)[gr.finger_bits] = gr.text;
}

yyerror(s)
     char *s;
{
  fprintf(stderr, "twini parser failed: line %d <%s>\n", curr_line, s);
  /* exit(1); */ /* now it can get handled... */
}
/* keycode : ZERO ',' number { $$.keyval = 0x80 + $3.number; } */




#include "lex-twini.c"
